/*
 *
 * Copyright (C) 2003 Sadysta ( http://sadysta.ath.cx )
 * Copyright (C) 2007 Tomasz Ostrowski ( http://ostry.w.szu.pl ) - modified from 1.8.9 source
 *
 * This file is part of Free GG.
 *
 * Free GG is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Free GG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Free GG; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "TransferThread.h"
#pragma package(smart_init)

#include <winsock.h>
#include <stdio.h>
#include <sys\stat.h>
#include <math.h>

#include "Transfer.h"
#include "Settings.h"
//---------------------------------------------------------------------------

__fastcall TTransferThread::TTransferThread(bool CreateSuspended)
        : TThread(CreateSuspended)
{
        I = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TTransferThread::_Progress()
{
        if (I == NULL)
        {
                I = TransferForm -> List -> Items -> Add();
                I -> Caption = _file_name;
                I -> SubItems -> Add (IntToStr((__int64)ceil (_transferred / 1000)) + " / " + IntToStr((__int64)ceil (_size / 1000)) + " KB");
                I -> SubItems -> Add ("???");
                I -> ImageIndex = _img_index;
        } else
        {
                I -> SubItems -> Strings[0] = IntToStr((__int64)ceil (_transferred / 1000)) + " / " + IntToStr((__int64)ceil (_size / 1000)) + " KB";
        }
}
//---------------------------------------------------------------------------
void TTransferThread::Progress (AnsiString FileName, long transferred, long size, int img_index)
{
        _file_name = FileName;
        _transferred = transferred;
        _size = size;
        _img_index = img_index;
		Synchronize (&_Progress);
}
//---------------------------------------------------------------------------
void __fastcall TTransferThread::Log (AnsiString s)
{
        _log_msg = s;
        Synchronize (&_Log);
}
//---------------------------------------------------------------------------
void __fastcall TTransferThread::_Log()
{
        TransferForm -> TransLog -> Lines -> Add (_log_msg);
}
//---------------------------------------------------------------------------
int read (int socket, char *buf, int size, int timeout = 10)
{
        TIMEVAL t;
        fd_set f;
        int r;
        DWORD dwTime = GetTickCount();

        do
        {
                t.tv_sec = timeout;
                t.tv_usec = 0;
                FD_ZERO (&f);
                FD_SET (socket, &f);
                r = select (0, &f, NULL, NULL, &t);
        }
        while (!(r == 1 && FD_ISSET (socket, &f)) && GetTickCount() - dwTime < (DWORD)timeout * 1000);

        return recv (socket, buf, size, 0);
/*        else
        if (GetTickCount() - dwTime < (DWORD)timeout * 1000)
                goto start;
        else
                return -2;*/
}
//---------------------------------------------------------------------------
void TTransferThread::SendFile (AnsiString FileName, int socket)
{
        struct
        {
                int have_file;
                int have_fileinfo;
                int dunno1;
                int dunno2;
        } gg_dcc1;
        file_info_struct finfo;
        FILE *f;
        struct stat statbuf;
        file_accept fa;
        file_header fh;
        char buf [4096];
        int _read;
        long overall = 0;

        gg_dcc1.have_file = GG_DCC_HAVE_FILE;
        gg_dcc1.have_fileinfo = GG_DCC_HAVE_FILEINFO;
        gg_dcc1.dunno1 = gg_dcc1.dunno2 = 0;

        send (socket, (char*)&gg_dcc1, sizeof(gg_dcc1), 0);

        f = fopen (FileName.c_str(), "rb");
        stat (FileName.c_str(), &statbuf);

        memset (&finfo, 0, sizeof(finfo));
        finfo.size_hdw = (int)(statbuf.st_size >> 4);
        finfo.size_ldw = (int)(statbuf.st_size);
        strcpy (finfo.file_name, ExtractFileName(FileName).c_str());

//        ShowMessage (IntToStr ( *(long*)&(finfo.size_hdw) ));

        send (socket, (char*)&finfo, sizeof(finfo), 0);

        read (socket, (char*)&fa, sizeof(fa), 60);

        Log ("Sending file " + FileName + " (" + IntToStr(statbuf.st_size) + " B) ...");

        while ((_read = fread (buf, 1, sizeof(buf), f)) > 0)
        {
                if (_read < sizeof(buf))
                        fh.type = GG_DCC_LAST_FILEHEADER;
                else
                        fh.type = GG_DCC_FILEHEADER;

                fh.chunk_size = _read;
                fh.dunno = 0;

                Log ("Header: " + IntToStr (fh.type) + " / " + IntToStr (fh.chunk_size) + " B");
//                Log ("Data: " + (AnsiString)buf);

                send (socket, (char*)&fh, sizeof(fh), 0);
                send (socket, buf, _read, 0);
                overall += _read;

                Progress (finfo.file_name, overall, statbuf.st_size, 6);
        }

        fclose (f);

}
//---------------------------------------------------------------------------
void __fastcall TTransferThread::Execute()
{
        Log ("TransferThread spawned.");

        sockaddr_in cli_addr;
        int cli_len = sizeof(cli_addr), n;
        char buf [128];
        int uin_from, uin_to, gg_dcc;
        file_info_struct file_info;
        file_accept fa;
        file_header fh;
        FILE *f;
        long file_size, overall;

        if (type == TT_ACTIVE_SEND_FILE)
        {
                Log ("Trying to active send file...");

                client_sock = socket (AF_INET, SOCK_STREAM, 0);
                cli_addr.sin_family = AF_INET;
                cli_addr.sin_port = htons (d_port);
                cli_addr.sin_addr.S_un.S_addr = inet_addr (d_addr.c_str());
                connect (client_sock, (sockaddr*)&cli_addr, cli_len);

                uin_from = StrToInt (SettingsForm -> Edit1 -> Text);
                uin_to = d_uin;

                send (client_sock, (char*)&uin_from, sizeof(int), 0);
                send (client_sock, (char*)&uin_to, sizeof(int), 0);

                read (client_sock, buf, 4); // UDAG

                gg_dcc = GG_DCC_CATCH_FILE;

                send (client_sock, (char*)&gg_dcc, sizeof(int), 0);

                TOpenDialog* od = new TOpenDialog(NULL);
                if (od -> Execute()) SendFile (od -> FileName, client_sock);
                delete od;
                goto quit;
        }

        if (type == TT_PASSIVE_GET_FILE || type == TT_ACTIVE_GET_FILE)
        {
                if (type == TT_PASSIVE_GET_FILE)
                {
                        client_sock = accept (TransferForm -> server_sock, (sockaddr*)&cli_addr, &cli_len);
                        TTransferThread *T = new TTransferThread (true);
                        T -> type = type;
                        T -> Resume();
                } else
                {
                        Log ("Ktos chce przeslac plik pasywnie. IP: " + d_addr + ":" + IntToStr(d_port));
                        client_sock = socket (AF_INET, SOCK_STREAM, 0);
                        cli_addr.sin_family = AF_INET;
                        cli_addr.sin_port = htons (d_port);
                        cli_addr.sin_addr.S_un.S_addr = inet_addr (d_addr.c_str());
                        connect (client_sock, (sockaddr*)&cli_addr, cli_len);
                }

                if (type == TT_PASSIVE_GET_FILE)
                {
                        read (client_sock, (char*)&uin_from, sizeof(int));
                        read (client_sock, (char*)&uin_to, sizeof(int));
                        d_uin = uin_from;

                        strcpy (buf, "UDAG");
                        send (client_sock, buf, 4, 0);

                        read (client_sock, (char*)&gg_dcc, sizeof(int));
                        if (gg_dcc == GG_DCC_CATCH_FILE) Log ("GG_DCC_CATCH_FILE"); else
                        if (gg_dcc == GG_DCC_WANT_FILE)
                        {
                                Log ("GG_DCC_WANT_FILE");
                                TOpenDialog* od = new TOpenDialog(NULL);
                                if (od -> Execute()) SendFile (od -> FileName, client_sock);
                                delete od;
                                goto quit;
                        }
                } else
                {
                        uin_from = StrToIntDef (SettingsForm -> Edit1 -> Text, 0);
                        send (client_sock, (char*)&uin_from, sizeof(int), 0);
                        uin_to = d_uin;
                        send (client_sock, (char*)&uin_to, sizeof(int), 0);

                        read (client_sock, buf, 4); // UDAG

                        gg_dcc = GG_DCC_WANT_FILE;
                        send (client_sock, (char*)&gg_dcc, sizeof(int), 0);
                }

                Log ("Klient o numerku " + IntToStr(d_uin) + " probuje wyslac do ciebie plik");

                read (client_sock, (char*)&gg_dcc, sizeof(int));
                if (gg_dcc == GG_DCC_HAVE_FILE) Log ("GG_DCC_HAVE_FILE"); else Log ("[2] Cos sie jebie");

                read (client_sock, (char*)&gg_dcc, sizeof(int));
                if (gg_dcc == GG_DCC_HAVE_FILEINFO) Log ("GG_DCC_HAVE_FILEINFO"); else Log ("[3] Cos sie jebie");

                read (client_sock, (char*)&gg_dcc, sizeof(int));
                read (client_sock, (char*)&gg_dcc, sizeof(int));

                read (client_sock, (char*)&file_info, sizeof(file_info));
                file_size = file_info.size_hdw;
                file_size <<= 8;
                file_size += file_info.size_ldw;
                Log ("Plik " + ((AnsiString)file_info.file_name) + " (" + IntToStr(file_size) + " B)");

                gg_dcc = MessageBox (NULL, ((AnsiString)("Osoba o ponizszych namiarch:\n\nUIN: " + IntToStr(d_uin) + "\nIP: " + inet_ntoa(cli_addr.sin_addr) + ":" + IntToStr(ntohs (cli_addr.sin_port)) + "\n\nchce wyslac ci nastepujacy plik:\n\nNazwa: " + file_info.file_name + "\nRozmiar: " + IntToStr(file_size) + " bajtow\n\nCzy chcesz go odebrac ?")).c_str(), "Transfer pliku", MB_YESNO | MB_ICONQUESTION | MB_SYSTEMMODAL);
                if (gg_dcc == IDNO)
                {
                        Log ("Zdecydowales sie nie odbierac pliku.");
                        goto quit;
                }

                fa.type = GG_DCC_GIMME_FILE;
                fa.start = 0;
                fa.dunno = 0;

                send (client_sock, (char*)&fa, sizeof(fa), 0);
        }

        f = fopen ( ( (AnsiString)(TransferForm -> Edit4 -> Text + "\\" + file_info.file_name) ).c_str(), "wb");

        Log ("Receiving ... ");
        overall = 0;

        //DWORD dwTime = GetTickCount();
        while (1)
        {
                int _read;

                _read = read (client_sock, (char*)&fh, sizeof(fh), 30);
                if (_read == sizeof(fh))
                        Log ("File Header: " + IntToStr(fh.type) + " / " + IntToStr(fh.chunk_size) + " B");
                else
                {
                        Log ("Nie moge odebrac naglowka: " + IntToStr(_read) + ".");
                        break;
                }

                int to_read = fh.chunk_size;
                _read = 1;
                if (to_read > 0 && to_read <= 65535)
                {
                        char *vbuf, *_vbuf;
                        _vbuf = vbuf = (char*)malloc (to_read);
                        while (to_read > 0)
                        {
                                _read = read (client_sock, vbuf, to_read, 30);
                                if (_read > 0)
                                {
                                        fwrite (vbuf, _read, 1, f);
                                        to_read -= _read;
                                        vbuf += _read;
                                        overall += _read;
                                } else Log ("[4] Cos sie jebie");
                        }
                        free (_vbuf);

                        Progress (file_info.file_name, overall, file_size, 5);
                } else
                {
                        Log ("Jakis pojebany chunk_size !!!");
                        break;
                }

                if (to_read != 0)
                {
                        Log ("Zepsute polaczenie: " + IntToStr(to_read) + ".");
                        break;
                } else
                if (fh.type == GG_DCC_LAST_FILEHEADER)
                {
                        Log ("Plik zostal odebrany.");
                        break;
                }
        }

        fclose (f);

quit:

        Log ("TransferThread DIE.");
        closesocket (client_sock);
}
//---------------------------------------------------------------------------
